from gpt_image.partition import Partition, PartitionEntryArray
from gpt_image.geometry import Geometry
import uuid

PART_NAME = "test-part"


def test_partition_init_blank():
    part = Partition()
    # everything should be zeroed
    assert part.type_guid.data == b"\x00" * 16
    assert part.partition_guid.data == b"\x00" * 16
    assert part.first_lba.data == b"\x00" * 8
    assert part.last_lba.data == b"\x00" * 8
    assert part.attribute_flags.data == b"\x00" * 8
    assert part.partition_name.data == b"\x00" * 72


def test_partition_init_real():
    part = Partition(PART_NAME, 2 * 1024)
    assert type(uuid.UUID(bytes=part.type_guid.data)) is uuid.UUID
    assert type(uuid.UUID(bytes=part.partition_guid.data)) is uuid.UUID
    assert PART_NAME in part.partition_name.data.decode("utf_16_le")
    assert part.size == 2 * 1024


def test_add():
    geo = Geometry(8 * 1024 * 1024)
    pa = PartitionEntryArray(geo)
    new_part = Partition(PART_NAME, 2 * 1024)
    pa.add(new_part)
    assert PART_NAME in pa.entries[0].partition_name.data.decode("utf_16_le")


def test_get_next_partition():
    geo = Geometry(8 * 1024 * 1024)
    pa = PartitionEntryArray(geo)
    next = pa._get_next_partition()
    assert next == 0
    new_part = Partition(PART_NAME, 2 * 1024)
    pa.add(new_part)
    next = pa._get_next_partition()
    assert next == 1



import pathlib

from gpt_image.geometry import Geometry
from gpt_image.partition import Partition
from gpt_image.table import ProtectiveMBR, Table


class Disk:
    """GPT disk

    A disk objects represents a new or existing GPT disk image.  If the file exists,
    it is assumed to be an existing GPT image. If it does not, a new file is created.

    Attributes:
        image_path: file image path (absolute or relative)
        size: disk image size in bytes
        sector_size: disk sector size. This should not be changed, changes to the
          layout should be done through the Partition alignment attribute
        fresh_disk: boolean create a new disk
    """

    def __init__(
        self,
        image_path: str,
        size: int = 0,
        sector_size: int = 512,
        *,
        fresh_disk: bool = False,
    ) -> None:
        """Init Disk with a file path and size in bytes"""
        self.image_path = pathlib.Path(image_path)
        self.name = self.image_path.name
        self.size = size
        self.sector_size = sector_size
        self.geometry = Geometry(self.size, self.sector_size)
        self.table = Table(self.geometry)
        if fresh_disk:
            self._create_disk()
        else:
            # @TODO: handle existing disk
            pass

    def _create_disk(self):
        """Create the disk image on Disk

        Creates the basic image structure at the specified path. This zeros
        the disk and writes the protective MBR.

        """
        self.image_path.touch(exist_ok=False)
        with open(self.image_path, "r+b") as f:
            # zero entire disk
            f.write(b"\x00" * self.size)
            f.seek(ProtectiveMBR.PROTECTIVE_MBR_START)
            f.write(self.table.protective_mbr.as_bytes())
            f.seek(ProtectiveMBR.DISK_SIGNATURE_START)
            f.write(self.table.protective_mbr.signature.data)

    def update_table(self):
        """Update the GPT table

        Writes the GPT header and partition tables to disk. Actions that
        happen before this are not written to disk.

        """
        self.table.update()
        with open(self.image_path, "r+b") as f:
            # write primary header
            f.seek(self.geometry.primary_header_byte)
            f.write(self.table.primary_header.as_bytes())

            # write primary partition table
            f.seek(self.geometry.primary_array_byte)
            f.write(self.table.partitions.as_bytes())

            # move to secondary header location and write
            f.seek(self.geometry.backup_header_byte)
            f.write(self.table.secondary_header.as_bytes())

            # write secondary partition table
            f.seek(self.geometry.backup_array_byte)
            f.write(self.table.partitions.as_bytes())

    def write_data(self, data: bytes, partition: Partition, offset: int = 0) -> None:
        """Write data to disk

        Args:
            data: data to write to partition. only bytes supported
            partition: Partition object to write data to
            offset: byte offset for writing data. The default is 0 but can be set to
                support custom offsets
        """
        if not type(data) is bytes:
            raise ValueError(f"data must be of type bytes. found type: {type(data)}")

        with open(self.image_path, "r+b") as f:
            start_lba = int.from_bytes(partition.first_lba.data, "little")
            start_byte = int(start_lba * self.sector_size)
            with open(self.image_path, "r+b") as f:
                f.seek(start_byte + offset)
                f.write(data)



import uuid

import pytest
from gpt_image.geometry import Geometry
from gpt_image.table import Header, ProtectiveMBR, Table

DISK_SIZE = 2 * 1024 * 1024  # 2 MB disk


@pytest.fixture
def new_geometry():
    return Geometry(DISK_SIZE)


def test_protective_mbr_init(new_geometry):
    geo = new_geometry
    pmbr = ProtectiveMBR(geo)
    assert type(pmbr.as_bytes()) == bytes
    pmbr_bytes = pmbr.as_bytes()
    assert pmbr_bytes[4:5] == b"\xEE"
    assert pmbr_bytes[8:12] == geo.primary_header_lba.to_bytes(4, "little")
    assert pmbr_bytes[12:16] == (geo.total_sectors - 1).to_bytes(4, "little")


def test_header_init_primary(new_geometry):
    geo = new_geometry
    head = Header(geo, uuid.uuid4())
    assert head.backup is False
    assert head.primary_header_lba.data == geo.primary_header_lba.to_bytes(8, "little")
    assert head.secondary_header_lba.data == geo.backup_header_lba.to_bytes(8, "little")
    assert head.partition_array_start.data == geo.primary_array_lba.to_bytes(
        8, "little"
    )


def test_header_init_backup(new_geometry):
    geo = new_geometry
    head = Header(geo, uuid.uuid4(), is_backup=True)
    assert head.backup is True
    # ensure the header LBA has been swapped
    assert head.primary_header_lba.data == geo.backup_header_lba.to_bytes(8, "little")
    assert head.secondary_header_lba.data == geo.primary_header_lba.to_bytes(
        8, "little"
    )
    assert head.partition_entry_start_byte == 0
    assert head.header_start_byte == int(32 * geo.sector_size)


def test_table_init(new_geometry):
    geo = new_geometry
    table = Table(geo)
    assert table.primary_header.backup is False
    assert table.secondary_header.backup is True


def test_checksum_partitions(new_geometry):
    table = Table(new_geometry)
    # partition checksum will be blank before when initialized
    assert table.primary_header.partition_array_crc.data == b"\x00" * 4
    table.checksum_partitions(table.primary_header)
    # test that the checksum is no longer zeroed
    assert table.primary_header.partition_array_crc.data != b"\x00" * 4


def test_checksum_header(new_geometry):
    table = Table(new_geometry)
    assert table.primary_header.header_crc.data == b"\x00" * 4
    table.checksum_header(table.primary_header)
    assert table.primary_header.header_crc.data != b"\x00" * 4


def test_update(new_geometry):
    table = Table(new_geometry)
    table.update()
    assert table.primary_header.header_crc != b"\x00" * 4
    assert table.secondary_header.header_crc != b"\x00" * 4
    assert table.primary_header.partition_array_crc != b"\x00" * 4
    assert table.secondary_header.partition_array_crc != b"\x00" * 4



import uuid

from gpt_image import disk
from gpt_image.partition import Partition


def test_e2e():
    # this should produce a disk that can be opened and validated with gdisk/sgdisk
    new_disk = disk.Disk("test-disk.raw", 2 * 1024 * 1024, fresh_disk=True)
    part1 = Partition("test-part", 2 * 1024, uuid.uuid4())

    new_disk.table.partitions.add(part1)
    new_disk.update_table()



import uuid

import pytest
from gpt_image.disk import Disk
from gpt_image.partition import Partition

BYTE_DATA = b"\x01\x02\x03\x04"
DISK_SIZE = 2 * 1024 * 1024
OFFSET = 32


@pytest.fixture
def new_image(tmp_path):
    image_size = DISK_SIZE
    image_name = tmp_path / "test.img"
    abs_path = image_name.resolve()
    disk = Disk(str(abs_path), image_size, fresh_disk=True)
    return disk


def test_disk_new(new_image: Disk):
    """Test creating a new disk image"""
    new_image.update_table()
    assert new_image.size == DISK_SIZE


def test_write_data_default(new_image: Disk):
    part = Partition("test-part", 1024, uuid.uuid4(), 1)
    new_image.table.partitions.add(part)
    new_image.update_table()
    new_image.write_data(BYTE_DATA, part)
    with open(new_image.image_path, "r+b") as f:
        f.seek(int.from_bytes(part.first_lba.data, "little") * new_image.sector_size)
        assert f.read(4) == BYTE_DATA


def test_write_data_offset(new_image: Disk):
    part = Partition("test-part", 1024, uuid.uuid4(), 1)
    new_image.table.partitions.add(part)
    new_image.update_table()
    new_image.write_data(BYTE_DATA, part, OFFSET)
    new_image.write_data(BYTE_DATA, part, OFFSET + OFFSET)
    with open(new_image.image_path, "r+b") as f:
        f.seek(
            int.from_bytes(part.first_lba.data, "little") * new_image.sector_size
            + OFFSET
        )
        assert f.read(4) == BYTE_DATA
        f.seek(
            int.from_bytes(part.first_lba.data, "little") * new_image.sector_size
            + OFFSET
            + OFFSET
        )
        assert f.read(4) == BYTE_DATA


def test_write_data_type_error(new_image: Disk):
    part = Partition("test-part", 1024, uuid.uuid4())
    with pytest.raises(ValueError):
        new_image.write_data(str("foobar"), part)  # type: ignore



from typing import Union


class Entry:
    def __init__(self, offset: int, length: int, data: Union[str, bytes, int]):
        self.offset = offset
        self.length = length
        assert type(data) in [str, int, bytes], "data must be of type str, int or bytes"

        self.data = self._convert_data(data)

    def _convert_data(self, data: Union[str, bytes, int]) -> bytes:
        """Convert incoming data to bytes

        Convenience class that converts data to bytes of the correct length

        """
        if isinstance(data, int):
            return (data).to_bytes(self.length, "little")
        if isinstance(data, str):
            b_string = bytes(data, encoding="utf_16_le")
            padded = b_string + bytes(self.length - len(b_string))
            return padded
        return data
